

;Offsets in RAM
.equ	tMinute = 0
.equ	tHour = 1
.equ	tWeekday = 2
.equ	tDay = 3
.equ	tMonth = 4
.equ	tYear = 5
.equ	tFlags = 6

;FLAGS:
.equ	timeDST = 0 ;daylight saving time
.equ	timeAnnounceDST = 1 ;announce start/end of DST at end of hour

;return values for check_dst
.equ	DST_NO = 0      ;currently DST
.equ	DST_YES = 1     ;currently not DST
.equ	DST_STARTS = 2  ;02:xx on last sunday in march (set clock to 03:xx)
.equ	DST_UNKNOWN = 3 ;02a:xx or 02b:xx on last sunday in october
.equ	DST_ENDS = 4    ;03:xx on last sunday in october (switch to 02b:xx if DST still active)

;===============================================================================

.cseg

inctime:
	;increment time by 1 minute (without modifying Y)
	ldd	temp, Y+tMinute ;minutes
	rcall	bcdinc
	std	Y+tMinute, temp
	cpi	temp, 0x60
	brlo	inctime_end
	ldi	temp, 0x00
	std	Y+tMinute, temp
	ldd	temp, Y+tHour ;hours
	rcall	bcdinc
	std	Y+tHour, temp
	rcall	check_dst
	cpi	temp, DST_STARTS
	breq	inctime_dst_switch_start
	cpi	temp, DST_ENDS
	breq	inctime_dst_switch_end
inctime_dst_end:
	ldd	temp, Y+tHour
	cpi	temp, 0x24
	brlo	inctime_end
	ldi	temp, 0x00
	std	Y+tHour, temp
	ldd	temp, Y+tWeekday ;weekday
	inc	temp
	sbrc	temp, 3
	ldi	temp, 1
	std	Y+tWeekday, temp
	rcall	calculate_month_length
	ldd	temp, Y+tDay ;day
	rcall	bcdinc
	std	Y+tDay, temp
	cp	temp2, temp
	brsh	inctime_end
	ldi	temp, 0x01
	std	Y+tDay, temp
	ldd	temp, Y+tMonth ;month
	rcall	bcdinc
	std	Y+tMonth, temp
	cpi	temp, 0x13
	brlo	inctime_end
	ldi	temp, 0x01
	std	Y+tMonth, temp
	ldd	temp, Y+tYear ;year
	rcall	bcdinc
	std	Y+tYear, temp
	cpi	temp, 0x9A
	brlo	inctime_end
	ldi	temp, 0x00
	std	Y+tYear, temp
inctime_end:
	ret
	
inctime_dst_switch_start:
	;DST starts
	ldi	temp, 0x03
	std	Y+tHour, temp
	ldd	temp2, Y+tFlags
	ori	temp2, 1<<timeDST
	std	Y+tFlags, temp2
	rjmp	inctime_dst_end

inctime_dst_switch_end:
	;DST ends
	ldd	temp2, Y+tFlags
	sbrs	temp2, timeDST
	rjmp	inctime_dst_end
	ldi	temp, 0x02
	std	Y+tHour, temp
	andi	temp2, ~(1<<timeDST)
	std	Y+tFlags, temp2
	rjmp	inctime_dst_end
	
;--------------------

check_dst:
	;check if current date has DST time (without modifying Y)
	;return values (in 'temp'):
	;  DST_YES     - currently DST
	;  DST_NO      - currently not DST
	;  DST_STARTS  - 02:xx on last sunday in march (set clock to 03:xx)
	;  DST_UNKNOWN - 02a:xx or 02b:xx on last sunday in october
	;  DST_ENDS    - 03:xx on last sunday in october (switch to 02b:xx if DST still active)
	ldd	temp, Y+tMonth
	cpi	temp, 0x03
	brlo	check_dst_no
	cpi	temp, 0x11
	brsh	check_dst_no
	cpi	temp, 0x04
	brlo	check_dst_start
	cpi	temp, 0x10
	brlo	check_dst_yes
	
	;october
	ldd	temp, Y+tDay
	cpi	temp, 0x25
	brlo	check_dst_yes
	;last 7 days in october
	rcall	bcd2bin
	ldd	temp2, Y+tWeekday
	cpi	temp2, 7
	breq	check_dst_end_sunday
	subi	temp2, -31
	sub	temp2, temp
	cpi	temp2, 7
	brlo	check_dst_no
	rjmp	check_dst_yes
check_dst_end_sunday:
	;last sunday in october: DST ends today
	ldd	temp, Y+tHour
	cpi	temp, 0x02
	brlo	check_dst_yes ;00:00 ~ 01:59
	breq	check_dst_unknown_announce ;02:00 ~ 02:59
	cpi	temp, 0x04
	brsh	check_dst_no ;04:00 ~ 23:59
	rjmp	check_dst_switch_end ;03:00 ~ 03:59
	
check_dst_start:
	;march
	ldd	temp, Y+tDay
	cpi	temp, 0x25
	brlo	check_dst_no
	;last 7 days in march
	rcall	bcd2bin
	ldd	temp2, Y+tWeekday
	cpi	temp2, 7
	breq	check_dst_start_sunday
	subi	temp2, -31
	sub	temp2, temp
	cpi	temp2, 7
	brlo	check_dst_yes
	rjmp	check_dst_no
check_dst_start_sunday:
	;last sunday in march: DST starts today
	ldd	temp, Y+tHour
	cpi	temp, 0x01
	brlo	check_dst_no ;00:00 ~ 00:59
	breq	check_dst_no_announce ;01:00 ~ 01:59
	cpi	temp, 0x02
	breq	check_dst_switch_start ;02:00 ~ 02:59
	rjmp	check_dst_yes ;03:00 ~ 23:59

check_dst_no:
	ldd	temp, Y+tFlags
	andi	temp, ~(1<<timeAnnounceDST)
	std	Y+tFlags, temp
	ldi	temp, DST_NO
	ret
	
check_dst_no_announce:
	ldd	temp, Y+tFlags
	ori	temp, 1<<timeAnnounceDST ;announce start of DST during 01:xx
	std	Y+tFlags, temp
	ldi	temp, DST_NO
	ret
	
check_dst_yes:
	ldd	temp, Y+tFlags
	andi	temp, ~(1<<timeAnnounceDST)
	std	Y+tFlags, temp
	ldi	temp, DST_YES
	ret
	
check_dst_unknown_announce:
	ldd	temp, Y+tFlags
	andi	temp, ~(1<<timeAnnounceDST) ;don't announce during 02b:xx
	sbrc	temp, timeDST
	ori	temp, 1<<timeAnnounceDST ;announce end of DST during 02a:xx
	std	Y+tFlags, temp
	ldi	temp, DST_UNKNOWN
	ret
	
check_dst_switch_start:
	ldd	temp, Y+tFlags
	andi	temp, ~(1<<timeAnnounceDST)
	std	Y+tFlags, temp
	ldi	temp, DST_STARTS
	ret
	
check_dst_switch_end:
	ldd	temp, Y+tFlags
	andi	temp, ~(1<<timeAnnounceDST)
	std	Y+tFlags, temp
	ldi	temp, DST_ENDS
	ret
	
;--------------------

calculate_month_length:
	;calculate month length (result goes into temp2)
	;month and year are taken from RAM (Y pointer)
	ldd	temp, Y+tMonth
	ldi	temp2, 0x30
	cpi	temp, 0x08
	brlo	PC+2
	inc	temp
	sbrc	temp, 0
	ldi	temp2, 0x31
	cpi	temp, 0x02
	brne	calculate_month_length_end
	ldd	temp, Y+tYear
	rcall	bcd2bin
	ldi	temp2, 0x28
	andi	temp, 0x03
	brne	calculate_month_length_end
	ldi	temp2, 0x29
calculate_month_length_end:
	ret

;--------------------

bcdinc:
	;increment packed BCD number
	subi	temp, -7
	brhc	PC+2
	subi	temp, 6
	ret
	
;--------------------

bcd2bin:
	;convert BCD value to binary (slow but small version)
	clr	temp2
bcd2bin_loop:
	cpi	temp, 0x10
	brlo	bcd2bin_end
	subi	temp, 0x10
	subi	temp2, -10
	rjmp	bcd2bin_loop
bcd2bin_end:
	add	temp, temp2
	ret
	
;bcd2bin:
;	mov	temp2, temp
;	andi	temp, 0x0F
;	swap	temp2
;	andi	temp2, 0x0F
;	lsl	temp2
;	add	temp, temp2
;	lsl	temp2
;	lsl	temp2
;	add	temp, temp2
;	ret
	
	