;LCD driver for HD44780, connected for 4-bit-mode

.equ	ARROW_LEFT = 0x7F
.equ	ARROW_RIGHT = 0x7E

.equ	LCD_CTRL_PORT = PORTD
.equ	LCD_DATA_PORT = PORTB
.equ	LCD_RS = 2
.equ	LCD_E = 3

;===============================================================================

.macro _lcd_goto
	;move cursor to (x, y), home position is (0, 0)
	ldi	temp, 0x80 + (@0) + (@1)*0x40
	rcall	lcd_cmd
.endmacro

.macro _lcd_chr
	;display single char on the LCD
	ldi	temp, @0
	rcall	lcd_chr
.endmacro

.macro _lcd_str
	;display string on the LCD
	rcall	lcd_str
	.db @0, 0
.endmacro

.macro _lcd_eestr
	;display string from EEPROM on the LCD
	ldi	ZL, @0
	rcall	lcd_eestr
.endmacro

;===============================================================================

;lcd_str:
;	;display string on the LCD
;	;  PARAMS: 0-terminated string
;	pop	ZH
;	pop	ZL
;	lsl	ZL
;	rol	ZH
;lcd_str_loop:
;	lpm	temp, Z+
;	cpi	temp, 0
;	breq	lcd_str_end
;	rcall	lcd_chr
;	rjmp	lcd_str_loop
;lcd_str_end:
;	adiw	ZH:ZL, 1
;	lsr	ZH
;	ror	ZL
;	ijmp ;return
	
;--------------------

lcd_eestr:
	;display string from EEPROM memory (0-terminated)
	rcall	ee_read
	cpi	temp, 0
	breq	lcd_eestr_end
	rcall	lcd_chr
	rjmp	lcd_eestr
lcd_eestr_end:
	ret
	
;--------------------

lcd_cmd:
	;send command to LCD
	cbi	LCD_CTRL_PORT, LCD_RS
	rjmp	lcd_write

;--------------------

lcd_bcd:
	;display packed BCD value on the LCD
	push	temp
	swap	temp
	rcall	lcd_bcd_sub
	pop	temp
lcd_bcd_sub:
	andi	temp, 0x0F
	ori	temp, '0'
	;(fall through to lcd_chr)
;--------------------
lcd_chr:
	;send data byte to LCD
	sbi	LCD_CTRL_PORT, LCD_RS
lcd_write:
	mov	temp2, temp
	;andi	temp2, 0xF0
	ori	temp2, 0x0F
	cli
	out	LCD_DATA_PORT, temp2
	sbi	LCD_CTRL_PORT, LCD_E
	lpm	w1, Z ;3 cycles delay
	cbi	LCD_CTRL_PORT, LCD_E
	sei
	swap	temp
	;andi	temp, 0xF0
lcd_write_8:
	ori	temp, 0x0F
	cli
	out	LCD_DATA_PORT, temp
	sbi	LCD_CTRL_PORT, LCD_E
	lpm	w1, Z ;3 cycles delay
	cbi	LCD_CTRL_PORT, LCD_E
	sei
	ldi	w1, 1 ;delay 100 us
	;(fall through to lcd_delay)
;--------------------
lcd_delay:
	;delay: temp * ~100us (@ 4.194304 MHz)
	ldi	w2, 139
lcd_delay_loop:
	dec	w2
	brne	lcd_delay_loop
	dec	w1
	brne	lcd_delay
	ret

;--------------------

lcd_init:
	cbi	LCD_CTRL_PORT, LCD_RS
	ldi	w1, 250 ;delay 25 ms
	rcall	lcd_delay
	ldi	temp, 0x30
	rcall	lcd_write_8
	ldi	w1, 50 ;delay 5 ms
	rcall	lcd_delay
	rcall	lcd_write_8
	rcall	lcd_write_8
	ldi	temp, 0x20 ;Function Set: 4 bit
	rcall	lcd_write_8
	ldi	temp, 0x2A ;Function Set: 4 bit, 2 lines, 5x7 font
	rcall	lcd_cmd
	ldi	temp, 0x0C ;Display on/off: on, no cursor
	rcall	lcd_cmd
	ldi	temp, 0x06 ;Entry Mode Set: cursor increment, no shift
	rcall	lcd_cmd
	ldi	temp, 0x02 ;Return Home
	rcall	lcd_cmd
	rcall	lcd_delay_2ms ;delay 2ms
	;(fall through to lcd_clear)
;--------------------
lcd_clear:
	;clear display
	ldi	temp, 0x01
	rcall	lcd_cmd
	;(fall through to lcd_delay2ms)
;--------------------
lcd_delay_2ms:
	ldi	w1, 20 ;delay 2 ms
	rjmp	lcd_delay ;combined "rcall lcd_delay" and "ret"
